﻿/***************************************************************************

Copyright (c) Microsoft Corporation. All rights reserved.
This code is licensed under the Visual Studio SDK license terms.
THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.

***************************************************************************/

using System;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.ComponentModel.Design;
using Microsoft.Win32;
using Microsoft.VisualStudio.Shell.Interop;
using Microsoft.VisualStudio.OLE.Interop;
using Microsoft.VisualStudio.Shell;
using IOleServiceProvider = Microsoft.VisualStudio.OLE.Interop.IServiceProvider;
using Microsoft.Samples.VisualStudio.IronPythonInterfaces;
using IronPython.Hosting;

namespace Microsoft.Samples.VisualStudio.IronPythonConsole
{
    [PackageRegistration(UseManagedResourcesOnly = true)]
    [DefaultRegistryRoot("Software\\Microsoft\\VisualStudio\\9.0")]
    //[InstalledProductRegistration(false, "#100", "#102", "1.0", IconResourceID = 0)]
    [ProvideService(typeof(IPythonEngineProvider))]
    [ProvideToolWindow(typeof(ConsoleWindow))]
    [ProvideMenuResource(1000, 1)]
    [ProvideLoadKey("Standard", "1.0", "IronPython Studio", "Microsoft Corporation", 1)]
    [Guid(GuidList.guidIronPythonConsolePkgString)]
    public sealed class PythonConsolePackage : Package
    {
        /// <summary>
        /// Default constructor of the package.
        /// Inside this method you can place any initialization code that does not require 
        /// any Visual Studio service, because at this point the package object is created but 
        /// not sited yet inside the Visual Studio environment. The place to do all the other 
        /// initialization is the Initialize method.
        /// </summary>
        public PythonConsolePackage()
        {
            Trace.WriteLine(string.Format(CultureInfo.CurrentCulture, "Entering constructor for: {0}", this.ToString()));
            // This package has to proffer the IronPython engine provider as a Visual Studio
            // service. Note that for performance reasons we don't actually create any object here,
            // but instead we register a callback function that will create the object the first
            // time this package will receive a request for the service.
            IServiceContainer container = this as IServiceContainer;
            ServiceCreatorCallback callback = new ServiceCreatorCallback(CreateService);
            container.AddService(typeof(IPythonEngineProvider), callback, true);
        }

        /// <summary>
        /// Initialization function for the package.
        /// When this function is called, the package is sited, so it is possible to use the standard
        /// Visual Studio services.
        /// </summary>
        protected override void Initialize()
        {
            // Always call the base implementation of Initialize.
            base.Initialize();

            // Add our command handlers for menu (commands must exist in the .vsct file)
            OleMenuCommandService mcs = GetService(typeof(IMenuCommandService)) as OleMenuCommandService;
            if (null == mcs)
            {
                // If it is not possible to get the command service, then there is nothing to do.
                return;
            }

            // Create the command for the tool window
            CommandID toolwndCommandID = new CommandID(GuidList.guidIronPythonConsoleCmdSet, (int)PkgCmdIDList.cmdidIronPythonConsole);
            MenuCommand menuToolWin = new MenuCommand(new EventHandler(ShowConsole), toolwndCommandID);
            mcs.AddCommand(menuToolWin);
        }

        /// <summary>
        /// This function is called the first time the service container implemented by this package
        /// receives a request for one of the services added with AddService; the goal of this
        /// function is to create an instance of the requested service.
        /// </summary>
        private object CreateService(IServiceContainer container, Type serviceType)
        {
            if (serviceType == typeof(IPythonEngineProvider))
            {
                // Get the service provider for this package.
                IOleServiceProvider serviceProvider = GetService(typeof(IOleServiceProvider)) as IOleServiceProvider;
                // Create the engine provider.
                return new IronPythonEngineProvider(serviceProvider);
            }
            return null;
        }

        /// <summary>
        /// This function is called when the user clicks the menu item that shows the 
        /// console window. 
        /// </summary>
        private void ShowConsole(object sender, EventArgs e)
        {
            // FindToolWindow will search for this window and, because the 'create' flag
            // is set to true, will create a new instance if it can not find it.
            ToolWindowPane pane = this.FindToolWindow(typeof(ConsoleWindow), 0, true);
            if (null == pane)
            {
                throw new COMException(Resources.CanNotCreateConsole);
            }
            IVsWindowFrame frame = pane.Frame as IVsWindowFrame;
            if (null == frame)
            {
                throw new COMException(Resources.CanNotCreateConsole);
            }
            Microsoft.VisualStudio.ErrorHandler.ThrowOnFailure(frame.Show());
        }

    }
}